import pkg_resources, os, json, yaml, shutil, glob
import jsonschema

def cwd():
    return os.getcwd()

def abs_path(*paths):
    return os.path.abspath(os.path.join(*paths))
    
def glob_to_root(path):
    if "*" in path:
        return os.path.dirname(path.split("*")[0])
    return os.path.dirname(path)

def recurse_path(path):
    if os.path.isfile(path):
        return [path]
    elif os.path.isdir(path):
        if "*" in path:
            return glob.iglob(path)
        else:
            return glob.iglob(os.path.join(path, "**/*"), recursive=True)

def path_ignored(path, ignore):
    for i in ignore:
        if i in path:
            return True
    return False

def is_threatspec_path(path):
    return os.path.isfile(os.path.join(path, "threatspec.yaml"))

def create_directories(paths):
    for path in paths:
        try:
            os.mkdir(path)
        except FileExistsError:
            pass

def write_json_pretty(data, *path):
    path = os.path.join(*path)
    with open(path, 'w') as fh:
        json.dump(data, fh, indent=2)

def read_json(*path):
    path = os.path.join(*path)
    with open(path) as fh:
        return json.load(fh)

def write_file(data, *path):
    path = os.path.join(*path)
    with open(path, 'w') as fh:
        fh.write(data)

def read_yaml(*path):
    path = os.path.join(*path)
    with open(path) as fh:
        return yaml.load(fh, Loader=yaml.SafeLoader)

def write_yaml(data, *path):
    path = os.path.join(*path)
    with open(path, 'w') as fh:
        fh.write(yaml.dump(data))

def validate_yaml_file(file_path, schema_file):
    schema_path = resolve_pkg_file(schema_file)
    try:
        jsonschema.validate(read_yaml(file_path), read_yaml(schema_path))
    except jsonschema.exceptions.ValidationError as e:
        return (False, str(e))
    return (True, None)
    
def resolve_pkg_file(*path):
    return pkg_resources.resource_filename("threatspec", os.path.join(*path))
    
def copy_pkg_file(src, dest):
    if os.path.isfile(dest):
        raise FileExistsError("File {} already exists.".format(dest))
    source = resolve_pkg_file(src)
    shutil.copyfile(source, dest)



from pprint import pprint
from typing import List, Dict
import uuid, re

class Source():
    def __init__(self, annotation: str, code: str, filename: str, line: int):
        self.annotation = annotation
        self.code = code
        self.filename = filename
        self.line = line
        
    def as_dict(self):
        return {
            "annotation": self.annotation,
            "code": self.code,
            "filename": self.filename,
            "line": self.line
        }

class Threat():
    def __init__(self, id: str, run_id: str, name: str, description: str = ""):
        self.id = id
        self.run_id = run_id
        self.name = name
        self.description = description

    def as_dict(self):
        return {
            "id": self.id,
            "run_id": self.run_id,
            "name": self.name,
            "description": self.description
        }
        
class Control():
    def __init__(self, id: str, run_id: str, name: str, description: str = ""):
        self.id = id
        self.run_id = run_id
        self.name = name
        self.description = description
        
    def as_dict(self):
        return {
            "id": self.id,
            "run_id": self.run_id,
            "name": self.name,
            "description": self.description
        }

class Component():
    def __init__(self, id: str, run_id: str, name: str, description: str = "", paths: List[str]=[]):
        self.id = id
        self.run_id = run_id
        self.name = name
        self.description = description
        self.paths = paths

    def as_dict(self):
        return {
            "id": self.id,
            "run_id": self.run_id,
            "name": self.name,
            "description": self.description,
            "paths": self.paths
        }

class Mitigation():
    def __init__(self, control: Control, threat: Threat, component: Component, source: Source):
        self.control = control
        self.threat = threat
        self.component = component
        self.source = source
        
    def as_dict(self):
        return {
            "control": self.control.as_dict(),
            "threat": self.threat.as_dict(),
            "component": self.component.as_dict(),
            "source": self.source.as_dict()
        }

class Acceptance():
    def __init__(self, threat: Threat, component: Component, details: str, source: Source):
        self.threat = threat
        self.component = component
        self.details = details
        self.source = source
        
    def as_dict(self):
        return {
            "threat": self.threat.as_dict(),
            "component": self.component.as_dict(),
            "details": self.details,
            "source": self.source.as_dict()
        }

class Transfer():
    def __init__(self, threat: Threat, source_component: Component, destination_component: Component, details: str, source: Source):
        self.threat = threat
        self.source_component = source_component
        self.destination_component = destination_component
        self.details = details
        self.source = source
        
    def as_dict(self):
        return {
            "threat": self.threat.as_dict(),
            "source_component": self.source_component.as_dict(),
            "destination_component": self.destination_component.as_dict(),
            "details": self.details,
            "source": self.source.as_dict()
        }

class Exposure():
    def __init__(self, threat: Threat, component: Component, details: str, source: Source):
        self.threat = threat
        self.component = component
        self.details = details
        self.source = source
    
    def as_dict(self):
        return {
            "threat": self.threat.as_dict(),
            "component": self.component.as_dict(),
            "details": self.details,
            "source": self.source.as_dict()
        }

class Connection():
    def __init__(self, source_component: Component, destination_component: Component, direction: str, details: str, source: Source):
        self.source_component = source_component
        self.destination_component = destination_component
        self.direction = direction
        self.details = details
        self.source = source
        
    def as_dict(self):
        return {
            "source_component": self.source_component.as_dict(),
            "destination_component": self.destination_component.as_dict(),
            "direction": self.direction,
            "details": self.details,
            "source": self.source.as_dict()
        }

class Review():
    def __init__(self, component: Component, details: str, source: Source):
        self.component = component
        self.details = details
        self.source = source

    def as_dict(self):
        return {
            "component": self.component.as_dict(),
            "details": self.details,
            "source": self.source.as_dict()
        }

class Test():
    def __init__(self, component: Component, control: Control, source: Source):
        self.component = component
        self.control = control
        self.source = source
        
    def as_dict(self):
        return {
            "component": self.component.as_dict(),
            "control": self.control.as_dict(),
            "source": self.source.as_dict()
        }

class Library():
    def parse(self, name):
        # Don't parse if all we have is an ID
        m = re.match(r'^#[a-z0-9_]+$', name, re.M)
        if m:
            return name

        # TODO - write tests then handle special global ids #client and #server
        m = re.match(r'(?P<name>[^()]+)(?:(?P<id>\(.*?\))(?P<description>.*)?)?', name, re.M | re.I)
        if m:
            match = m.groupdict()

            if match["name"].startswith("#"):
                match["id"] = match["name"]

            if not match["id"]:
                if match["name"].endswith("/"): # Dirty hack
                    id_body = match["name"]+"root"
                else:
                    id_body = match["name"]
                match["id"] = "#" + re.sub('[^a-z0-9_]+', '_', id_body.strip().lower().replace('-','')).strip('_')

            if match["id"][0] == "(" and match["id"][-1] == ")":
                match["id"] = match["id"][1:-1]

            if not match["id"].startswith("#"): # Very, very dirty hack
                if match["id"].endswith("/"): # Dirty hack
                    id_body = match["id"]+"root"
                else:
                    id_body = match["id"]
                match["id"] = "#" + re.sub('[^a-z0-9_]+', '_', id_body.strip().lower().replace('-','')).strip('_')

            if not match["description"]:
                match["description"] = ""
            return match
        else:
            raise RuntimeError("Failed to parse ID: {}".format(name))

class ThreatLibrary(Library):
    def __init__(self, threats: Dict[str, Threat] = {}):
        self.threats = threats

    def add_threat(self, name=None, run_id=None):
        data = self.parse(name)
        if isinstance(data, str):
            return data
        if not data["id"] in self.threats:
            self.threats[data["id"]] = Threat(data["id"], run_id, data["name"], data["description"]) # TODO: Handle id clash
        return data["id"]

    def load(self, data):
        for id, threat in data["threats"].items():
            if id not in self.threats: # TODO Handle id clash
                self.threats[id] = Threat(id, threat["run_id"], threat["name"], threat["description"])
                
    def save(self, run_id=None):
        if not run_id:
            return [t.as_dict() for t in self.threats]
            
        data = {"threats":{}}
        for id, threat in self.threats.items():
            if not threat.run_id:
                continue
            if threat.run_id == run_id:
                data["threats"][id] = threat.as_dict()
        return data
        
class ControlLibrary(Library):
    def __init__(self, controls: Dict[str, Control] = {}):
        self.controls = controls

    def add_control(self, name=None, run_id=None):
        data = self.parse(name)
        if isinstance(data, str):
            return data
        if not data["id"] in self.controls:
            self.controls[data["id"]] = Control(data["id"], run_id, data["name"], data["description"])
        return data["id"]

    def load(self, data):
        for id, control in data["controls"].items():
            if id not in self.controls: # TODO Handle id clash
                self.controls[id] = Control(id, control["run_id"], control["name"], control["description"])
    
    def save(self, run_id=None):
        if not run_id:
            return [c.as_dict() for c in self.controls]
        
        data = {"controls":{}}
        for id, control in self.controls.items():
            if not control.run_id:
                continue
            if control.run_id == run_id:
                data["controls"][id] = control.as_dict()
        return data

class ComponentLibrary(Library):
    def __init__(self, components: Dict[str, Component] = {}):
        self.components = components

    def add_component(self, name=None, run_id=None):
        data = self.parse(name)
        if isinstance(data, str):
            return data

        path = data["name"].split(":")[0:-1] # Ignore the last one as that's the component itself

        if not data["id"] in self.components:
            self.components[data["id"]] = Component(data["id"], run_id, data["name"], data["description"])
        elif path not in self.components[data["id"]].paths:
            self.components[data["id"]].paths.append(path)
        return data["id"]

    def load(self, data):
        for id, component in data["components"].items():
            if id not in self.components: # TODO Handle id clash
                self.components[id] = Component(id, component["run_id"], component["name"], component["description"], component["paths"])
                
    def save(self, run_id=None):
        if not run_id:
            return [c.as_dict() for c in self.components]
        
        data = {"components":{}}
        for id, component in self.components.items():
            if not component.run_id:
                continue
            if component.run_id == run_id:
                data["components"][id] = component.as_dict()
        return data
                

class ThreatModel(Library):
    def __init__(self,
        mitigations: List[Mitigation] = [],
        acceptances: List[Acceptance] = [],
        transfers: List[Transfer] = [],
        exposures: List[Exposure] = [],
        connections: List[Connection] = [],
        reviews: List[Review] = [],
        tests: List[Test] = [],
        run_id: str = ""):
        
        self.mitigations = mitigations
        self.acceptances = acceptances
        self.transfers = transfers
        self.exposures = exposures
        self.connections = connections
        self.reviews = reviews
        self.tests = tests
        self.run_id = run_id
        
        self.threat_library = None
        self.control_library = None
        self.component_library =  None


    def add_mitigation(self, threat, control, component, source):
        self.mitigations.append(Mitigation(
            self.control_library.add_control(control, self.run_id),
            self.threat_library.add_threat(threat, self.run_id),
            self.component_library.add_component(component, self.run_id),
            Source(**source)
        ))

    def add_acceptance(self, threat, component, details, source):
        self.acceptances.append(Acceptance(
            self.threat_library.add_threat(threat, self.run_id),
            self.component_library.add_component(component, self.run_id),
            details,
            Source(**source)
        ))

    def add_transfer(self, threat, source_component, destination_component, details, source):
        self.transfers.append(Transfer(
            self.threat_library.add_threat(threat, self.run_id),
            self.component_library.add_component(source_component, self.run_id),
            self.component_library.add_component(destination_component, self.run_id),
            details,
            Source(**source)
        ))

    def add_exposure(self, threat, component, details, source):
        self.exposures.append(Exposure(
            self.threat_library.add_threat(threat, self.run_id),
            self.component_library.add_component(component, self.run_id),
            details,
            Source(**source)
        ))

    def add_connection(self, source_component, destination_component, direction, details, source):
        self.connections.append(Connection(
            self.component_library.add_component(source_component, self.run_id),
            self.component_library.add_component(destination_component, self.run_id),
            direction,
            details,
            Source(**source)
            ))

    def add_review(self, component, details, source):
        self.reviews.append(Review(
            self.component_library.add_component(component, self.run_id),
            details,
            Source(**source)
        ))

    def add_test(self, component, control, source):
        self.tests.append(Test(
            self.component_library.add_component(component, self.run_id),
            self.control_library.add_control(control, self.run_id),
            Source(**source)
        ))

    def load(self, data):
        for mitigation in data["mitigations"]:
            self.add_mitigation(**mitigation)
        for exposure in data["exposures"]:
            self.add_exposure(**exposure)
        for transfer in data["transfers"]:
            self.add_transfer(**transfer)
        for acceptance in data["acceptances"]:
            self.add_acceptance(**acceptance)
        for connection in data["connections"]:
            self.add_connection(**connection)
        for review in data["reviews"]:
            self.add_review(**review)
        for test in data["tests"]:
            self.add_test(**test)
            
    def save(self):
        return {
            "mitigations": self.mitigations,
            "exposures": self.exposures,
            "transfers": self.transfers,
            "acceptances": self.acceptances,
            "connections": self.connections,
            "reviews": self.reviews,
            "tests": self.tests
        }


import logging
logger = logging.getLogger(__name__)

import click, sys
from threatspec import app

def validate_logging(ctx, param, value):
    levels = {
        "none": 100,
        "crit": logging.CRITICAL,
        "error": logging.ERROR,
        "warn": logging.WARNING,
        "info": logging.INFO,
        "debug": logging.DEBUG
    }
    if value.lower() in levels:
        return levels[value.lower()]
    raise click.BadParameter("Log level must be one of: {}".format(", ".join(levels.keys())))

def configure_logger(level, verbose):
    if verbose:
        logging.basicConfig(format='%(asctime)s %(levelname)s: %(message)s', level=level)
    else:
        logging.basicConfig(format='%(message)s', level=level)

@click.group()
@click.option("--log-level", "-l", callback=validate_logging, default="info", help="Set the log level. Must be one of: crit, error, warn, info, debug, none.")
@click.option("--verbose/--no-verbose", default=False, help="Makes logging more verbose.")
@click.version_option()
def cli(log_level, verbose):
    """
    threatspec - threat modeling as code

    threatspec is an open source project that aims to close the gap between
    development and security by bringing the threat modelling process further
    into the development process. This is achieved by having developers and
    security engineers write threat specifications alongside code, then
    dynamically generating reports and data-flow diagrams from the code. This
    allows engineers to capture the security context of the code they write,
    as they write it.
    
    Usage:
    
    # Initialise threatspec in the current directory
    $ threatspec init
    
    # Configure the source code paths
    $ $EDITOR threatspec.yaml
    
    # Run threatspec against the source code paths
    $ threatspec run
    
    # Generate the threat mode report
    $ threatspec report
    
    For more information for each subcommand use --help. For everything else,
    visit the website at https://threatspec.org
    """

    configure_logger(log_level, verbose)
    
@cli.command()
def init():
    """
    Initialise threatspec in the current directory.

    This will create a project configuration file called threatspec.yaml. Edit
    this file to configure the project name and description as well the source
    code paths for threatspec to scan.

    This command will also create the threatmodel directory in the current
    path. This directory contains the json output files from threatspec run.

    The following file contains the collection of mitigations, acceptances,
    connections etc identified as annotations in code:

        threatmodel/threatmodel.json

    The following three threat model library files are loaded each time threatspec
    is run. If new threats, controls or components are found, they are added to these
    files.
    
    This allows threats, controls and components to be used across projects
    and allows you to create threat library files, for example from OWASP or CWE
    data. When threatspec loads paths configured in threatspec.yaml, it checks
    each path to see if a threatspec.yaml file exists. If so, it attempts to load the
    below files.

    threatmodel/threats.json threatmodel/controls.json threatmodel/components.json
    """

    threatspec = app.ThreatSpecApp()
    threatspec.init()

@cli.command()
def run():
    """
    Run threatspec against source code files.

    This command loads the configuration file and for each configured path it first
    checks to see if a threatspec.yaml file exists in the path. If it does, it loads
    the three library json files.

    Once all the library files have been loaded from the paths, threatspec run will
    recursively parse each file in the path, looking for threatspec annotations.
    
    You can exclude patterns from being searched (for example 'node_modules') using the
    'ignore' key for the paths in the configuration file. See the documentation for
    more information.

    After all the source files have parsed, threatspec run will generate the
    threatmodel/threatmodel.json file as well as the three library files:

    threatmodel/threats.json threatmodel/controls.json threatmodel/components.json
    """

    threatspec = app.ThreatSpecApp()
    threatspec.run()

@cli.command()
def report():
    """
    Generate the threatspec threat model report.

    This will use Graphviz to generate a visualisation of the threat model, and
    embed it in a threat model markdown document in the current directory:
    
    ThreatModel.md

    This document contains tables of mitigations etc (including any tests), as
    well as connections and reviews.
    """

    threatspec = app.ThreatSpecApp()
    threatspec.report()
    
if __name__ == '__main__':
    cli(None, None)
